package jpos.config.simple.xml;

///////////////////////////////////////////////////////////////////////////////
//
// This software is provided "AS IS".  The JavaPOS working group (including
// each of the Corporate members, contributors and individuals)  MAKES NO
// REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF THE SOFTWARE,
// EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE IMPLIED 
// WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND 
// NON-INFRINGEMENT. The JavaPOS working group shall not be liable for
// any damages suffered as a result of using, modifying or distributing this
// software or its derivatives. Permission to use, copy, modify, and distribute
// the software and its documentation for any purpose is hereby granted. 
//
///////////////////////////////////////////////////////////////////////////////

import java.io.*;
import java.util.*;
import java.net.URL;
import java.text.DateFormat;

import java.util.*;
import java.io.FileInputStream;
import java.io.PrintWriter;

import org.w3c.dom.*;
                  
import org.apache.xerces.parsers.DOMParser;

import org.apache.xerces.dom.DocumentImpl;
import org.apache.xerces.dom.DocumentTypeImpl;
import org.apache.xerces.dom.DOMImplementationImpl;

import org.apache.xml.serialize.XMLSerializer;
import org.apache.xml.serialize.OutputFormat;

import org.xml.sax.InputSource;

import jpos.config.*;
import jpos.config.simple.*;
import jpos.util.*;
import jpos.loader.Version;

/**
 * Simple implementation of the JposRegPopulator that loads and saves the entries
 * in XML using the "jpos/res/jcl.dtd" DTD and the XML4J (Xerces) API
 * NOTE: this class must define a public no-argument ctor so that it may be created 
 *       via reflection when its defined in the jpos.properties as 
 *       the jpos.config.regPopulatorClass
 * @see jpos.util.JposProperties#JPOS_REG_POPULATOR_CLASS_PROP_NAME
 * @since 1.2 (NY 2K meeting)
 * @author E. Michael Maximilien (maxim@us.ibm.com)
 */
public class XercesRegPopulator extends AbstractRegPopulator implements XmlRegPopulator
{
    //-------------------------------------------------------------------------
    // Ctor(s)
    //

    /**
     * Default ctor
     * @since 1.2 (NY 2K meeting)
     */
    public XercesRegPopulator() { super( XercesRegPopulator.class.getName() ); }

    /**
     * 1-arg constructor that takes the unique ID
	 * @param s the unique ID string
     * @since 1.3 (Washington DC 2001)
     */
    public XercesRegPopulator( String s ) { super( s ); }

    //-------------------------------------------------------------------------
    // Public methods
    //

	/**
	 * @return the fully qualified class name implementing the JposRegPopulator interface
	 * @since 1.3 (Washington DC 2001 meeting)
	 */
	public String getClassName() { return XercesRegPopulator.class.getName(); }

    /**
     * Tell the populator to save the current entries 
     * @param entries an enumeration of JposEntry objects
     * @since 1.2 (NY 2K meeting)
     * @throws java.lang.Exception if any error occurs while saving
     */
    public void save( Enumeration entries ) throws Exception
    {
        if( isPopulatorFileDefined() )
            convertJposEntriesToXml( entries, getPopulatorFileOS() );
        else
            convertJposEntriesToXml( entries, new FileOutputStream( xmlFileName ) );
    }

    /**
     * Tell the populator to save the current entries in the file specified 
     * @param entries an enumeration of JposEntry objects
     * @param xmlFileName the XML file name to save entries
     * @since 1.3 (SF 2K meeting)
     * @throws java.lang.Exception if any error occurs while saving
     */
    public void save( Enumeration entries, String xmlFileName ) throws Exception
    { 
		File xmlFile = new File( xmlFileName );
		FileOutputStream fos = new FileOutputStream( xmlFile );

		convertJposEntriesToXml( entries, fos ); 

		fos.close();
	}

    /**
     * Tell the populator to load the entries 
     * @since 1.2 (NY 2K meeting)
     */
    public void load()
    {
        if( isPopulatorFileDefined() == false )
        {
            getJposEntries().clear();
            xmlFileName = DEFAULT_XML_FILE_NAME;
			load( xmlFileName );

            return;
        }

        try
        {
            getJposEntries().clear();

            checkAndCreateTempDTD();

            domParser.parse( new InputSource( getPopulatorFileIS() ) );

            Document document = domParser.getDocument();

            Enumeration entries = extractJposEntries( document );

            while( entries.hasMoreElements() )
            {
                JposEntry jposEntry = (JposEntry)entries.nextElement();

                if( jposEntry.hasPropertyWithName( JposEntry.LOGICAL_NAME_PROP_NAME ) )
					getJposEntries().put( jposEntry.getPropertyValue( JposEntry.LOGICAL_NAME_PROP_NAME ),
                                          jposEntry );
            }

			lastLoadException = null;
        }
        catch( Exception e )
        {
			lastLoadException = e;
			tracerPrintln( "Error loading XML file.  Error message = " + e.getMessage() ); 
		}
        finally
        { removeTempDTD(); }
    }

    /**
     * Loads the entries specified in the xmlFileName
     * NOTE: this method is for testing purposes (use public load() method)
     * @param xmlFileName the XML file name
     * @since 1.3 (SF 2K meeting)
     */
    public void load( String xmlFileName )
    {
		InputStream is = null;
		File xmlFile = new File( xmlFileName );

		try
		{
			if( xmlFile.exists() )
				is = new FileInputStream( xmlFile );
			else
				is = findFileInClasspath( xmlFileName );

			if( is == null )
			{
				getJposEntries().clear();

				String message = "Could not find file: " + xmlFileName + " in path or CLASSPATH";
				tracerPrintln( message );

				lastLoadException = new FileNotFoundException( xmlFileName );
				
				return;
			}

			lastLoadException = null;
		}
		catch( Exception e )
		{
			lastLoadException = e;
			tracerPrintln( "Error loading XML file.  Exception.message = " + e.getMessage() ); 
		}

        try
        {
			getJposEntries().clear();

            checkAndCreateTempDTD();

            domParser.parse( new InputSource( is ) );
            
            Document document = domParser.getDocument();

            Enumeration entries = extractJposEntries( document );

            while( entries.hasMoreElements() )
            {
                JposEntry jposEntry = (JposEntry)entries.nextElement();

                if( jposEntry.hasPropertyWithName( JposEntry.LOGICAL_NAME_PROP_NAME ) )
                    getJposEntries().put( jposEntry.getPropertyValue( JposEntry.LOGICAL_NAME_PROP_NAME ),
                                          jposEntry );
            }

			lastLoadException = null;

        }
        catch( Exception e )
        {
			lastLoadException = e;
			tracerPrintln( "Error loading XML file.  Exception.message = " + e.getMessage() ); 
		}
		finally
        { removeTempDTD(); }
    }

    /**
     * @return the URL pointing to the entries file loaded or saved
     * @since 1.2 (NY 2K meeting)
     */
    public URL getEntriesURL()
    {
        URL url = null;

        if( getPopulatorFileURL() != null && !getPopulatorFileURL().equals( "" ) )
            try{ url =  new URL( getPopulatorFileURL() ); }
            catch( Exception e ) {}
        else
            url = createURLFromFile( new File( getPopulatorFileName() ) );

        return url;
    }

	/**
	 * @return the name of this populator.  This should be a short descriptive name
	 * @since 1.3 (Washington DC 2001 meeting)
	 */
	public String getName() { return XERCES_REG_POPULATOR_NAME_STRING; }

    //--------------------------------------------------------------------------
    // Private methods
    //

	/** @return the Tracer object */
	private Tracer getTracer() { return Tracer.getInstance(); }

	/** 
	 * Println the message to the Tracer
	 * @param message the String message
	 */
	private void tracerPrintln( String message ) { getTracer().println( message ); }

    //--------------------------------------------------------------------------
    // Protected methods
    //

    /**
     * Checks if JCL DTD is available "jpos/res/jcl.dtd" if not then creates it
     * extracting it from the JAR file
     * @since 1.2 (SF 2K meeting)
     */
    protected void checkAndCreateTempDTD()
    {
        createdTempDTD = false;
        createdTempDir = false;
        InputStream is = null;

        try
        {
            File dtdPath = new File( DTD_FILE_PATH );
            File dtdFile = new File( DTD_FILE_NAME );

            if( dtdFile.exists() )
				return;            
            
            if( !dtdPath.exists() )
            {
                dtdPath.mkdirs();
                createdTempDir = true;
				tracerPrintln( "DTD file PATH does not exist...  created jpos/res" );
            }
            
            //Extract from JAR file and create...
            is = ClassLoader.getSystemResourceAsStream( DTD_JAR_FILE_NAME );

			if( is == null )
				is = ClassLoader.getSystemClassLoader().getResourceAsStream( DTD_FILE_NAME );

			tracerPrintln( "Got DTD InputStream from current ClassLoader" );

            if( is != null )
				readAndCreateTempDtdFile( is );
        }
        catch( IOException ioe )
        { tracerPrintln( "Error creating DTD file.  Exception.message = " + ioe.getMessage() ); }
        finally
        {
            try{ if( is != null ) is.close(); }
            catch( IOException ioe ) {}
        }
    }

	/**
	 * Reads the DTD file from the InputStream provided and creates a temp
	 * @since 1.2 (SF 2K meeting)
	 * @exception java.io.IOException if anything goes wrong
	 */
	protected void readAndCreateTempDtdFile( InputStream is ) throws IOException
	{
		File dtdFile = new File( DTD_FILE_NAME );
		FileOutputStream fos = new FileOutputStream( dtdFile );
		OutputStreamWriter osw = new OutputStreamWriter( fos );
		BufferedInputStream bis = new BufferedInputStream( is );

		StringBuffer sb = new StringBuffer();

		while( bis.available() > 0 )
		{
			byte[] buffer = new byte[ bis.available() ];

			bis.read( buffer );

			sb.append( new String( buffer ) );
		}

		osw.write( sb.toString().trim() );

		createdTempDTD = true;

		try
		{ 
			if( osw != null ) osw.close(); 
			if( fos != null ) fos.close(); 
		}
		catch( IOException ioe ) {}

		tracerPrintln( "Read and created temp DTD file jpos/res/jcl.dtd" );
	}

    /**
     * Removes JCL DTD "jpos/res/jcl.dtd" if it was created 
     * @since 1.2 (SF 2K meeting)
     */
    protected void removeTempDTD()
    {
        try
        {
            if( createdTempDTD )
            {
                File dtdFile = new File( DTD_FILE_NAME );
                dtdFile.delete();
            }
            
            if( createdTempDir )
            {
				tracerPrintln( "Deleting the DTD temp jpos/res/ directories..." );

                File dtdJposResFilePath = new File( DTD_JPOS_RES_FILE_PATH );
                dtdJposResFilePath.delete();

                File dtdJposFilePath = new File( DTD_JPOS_FILE_PATH );
                dtdJposFilePath.delete();
            }
        }
        catch( Exception e )
        { tracerPrintln( "Error removing temporary DTD file.  Exception.message = " + e.getMessage() ); }
    }

    /**
     * @return the DOM parser object
     * @since 1.2 (NY 2K meeting)
     */
    protected DOMParser getParser() { return domParser; }

    /**
     * @return an enumeration of JposEntry objects read from the XML document object
     * @param document the XML document object
     * @since 1.2 (NY 2K meeting)
     */
    protected Enumeration extractJposEntries( Document document )
    {
        Vector entries = new Vector();

		NodeList nodeList = document.getElementsByTagName( "JposEntry" );

		String currentEntryLogicalName = "";

		try
		{
			for( int i = 0; i < nodeList.getLength(); ++i )
			{
				Node node = nodeList.item( i );

				if( node.getNodeType() != Node.ELEMENT_NODE )
					continue;

				JposEntry jposEntry = new SimpleEntry();

				Element jposEntryElement = (Element)node;

				currentEntryLogicalName = jposEntryElement.getAttribute( "logicalName" );
				jposEntry.addProperty( "logicalName", currentEntryLogicalName );

				NodeList childList = nodeList.item( i ).getChildNodes();

				for( int j = 0; j < childList.getLength(); ++j )
				{
					Node child = childList.item( j );

					if( child.getNodeType() != Node.ELEMENT_NODE )
						continue;

					Element element = (Element)child;

					String elementName = element.getNodeName();

					if( elementName.equals( "creation" ) )
						extractCreationAttr( jposEntry, element );
					else
					if( elementName.equals( "vendor" ) )
						extractVendorAttr( jposEntry, element );
					else
					if( elementName.equals( "jpos" ) )
						extractJposAttr( jposEntry, element );
					else
					if( elementName.equals( "product" ) )
						extractProductAttr( jposEntry, element );
					else
						extractPropAttr( jposEntry, element );
				}

				if( JposEntryUtility.isValidJposEntry( jposEntry ) )
					entries.addElement( jposEntry );
				else
					throw new JposConfigException( "JposEntry with logicalName: " + currentEntryLogicalName + " is not valid (missing required properties)" );
			}
		}
		catch( JposConfigException jce )
		{
			tracerPrintln( "Skipping invalid entry with logicalName: " + currentEntryLogicalName );
			tracerPrintln( "--->Exception.message = " + jce.getMessage() );

			jce.printStackTrace();
			if( jce.getOrigException() != null ) jce.getOrigException().printStackTrace();
		}

        return entries.elements();
    }

    /**
     * Get the <creation> element attributes and adds corresponding properties to JposEntry
     * @param jposEntry the entry to add properties to
     * @param element the <creation> XML element
     * @since 1.2 (NY 2K meeting)
     */
    protected void extractCreationAttr( JposEntry jposEntry, Element element )
    {
        jposEntry.addProperty( "serviceInstanceFactoryClass", element.getAttribute( "factoryClass" ) );
        jposEntry.addProperty( "serviceClass", element.getAttribute( "serviceClass" ) );
    }

    /**
     * Get the <vendor> element attributes and adds corresponding properties to JposEntry
     * @param jposEntry the entry to add properties to
     * @param element the <vendor> XML element
     * @since 1.2 (NY 2K meeting)
     */
    protected void extractVendorAttr( JposEntry jposEntry, Element element )
    {
        jposEntry.addProperty( "vendorName", element.getAttribute( "name" ) );
        jposEntry.addProperty( "vendorURL", element.getAttribute( "url" ) );
    }

    /**
     * Get the <jpos> element attributes and adds corresponding properties to JposEntry
     * @param jposEntry the entry to add properties to
     * @param element the <jpos> XML element
     * @since 1.2 (NY 2K meeting)
     */
    protected void extractJposAttr( JposEntry jposEntry, Element element )
    {
        jposEntry.addProperty( "jposVersion", element.getAttribute( "version" ) );
        jposEntry.addProperty( "deviceCategory", element.getAttribute( "category" ) );
    }

    /**
     * Get the <product> element attributes and adds corresponding properties to JposEntry
     * @param jposEntry the entry to add properties to
     * @param element the <product> XML element
     * @since 1.2 (NY 2K meeting)
     */
    protected void extractProductAttr( JposEntry jposEntry, Element element )
    {
        jposEntry.addProperty( "productName", element.getAttribute( "name" ) );
        jposEntry.addProperty( "productDescription", element.getAttribute( "description" ) );
        jposEntry.addProperty( "productURL", element.getAttribute( "url" ) );
    }

    /**
     * Get the <prop> element attributes and adds corresponding properties to JposEntry
     * @param jposEntry the entry to add properties to
     * @param element the <prop> XML element
     * @since 1.2 (NY 2K meeting)
	 * @throws jpos.config.JposConfigException if the property value does not match the type or is
	 * not a valid value (like for instance an invalid number)
     */
    protected void extractPropAttr( JposEntry jposEntry, Element element ) throws JposConfigException
    {
		String propName = element.getAttribute( "name" );
		String propValueString = element.getAttribute( "value" );
		String propTypeString = element.getAttribute( "type" );

		if( propTypeString.equals( "" ) ) propTypeString = "String";

		Object propValue = null;
		Class propType = null;

		try
		{
			propType = Class.forName( ( propTypeString.startsWith( "java.lang." ) ? propTypeString : 
				                                                                    "java.lang." + propTypeString ) ); 
		}
		catch( ClassNotFoundException cnfe )
		{ throw new JposConfigException( "Invalid property type: " + propTypeString + " for property named: " + propName , cnfe ); }

		if( JposEntryUtility.isValidPropType( propType ) == false )
			throw new JposConfigException( "Invalid property type: " + propTypeString + " for property named: " + propName );

		propValue = JposEntryUtility.parsePropValue( propValueString, propType );

		if( JposEntryUtility.validatePropValue( propValue, propType ) == false )
			throw new JposConfigException( "Invalid property type: " + propTypeString + " for property named: " + propName );

        jposEntry.add( jposEntry.createProp( propName, propValue, propType ) );
    }

    /**
     * Converts an Enumeration of JposEntry objects to XML
     * @param entries an Enumeration of JposEntry objects
     * @param os the OutputStream to stream the entries to
     * @exception java.lang.Exception if something goes wrong serializing
     * @since 1.2 (NY 2K meeting)
     */
    protected void convertJposEntriesToXml( Enumeration entries,
                                            OutputStream os ) throws Exception
    {
        Document document = getParser().getDocument();
        serializeDocument( document, entries, os );
    }

    /**
     * Serializes the JposEntry objects to an XML document and save to OutputStream
     * @param document the XML document object
     * @param entries an Enumeration of JposEntry objects
     * @param os the OuputStream object
     * @exception java.lang.Exception anything goes wrong while saving
     * @since 1.2 (NY 2K meeting)
     */
    protected void serializeDocument( Document document, Enumeration entries, OutputStream os ) throws Exception
    {
        Document newDoc = createEmptyDocument();

        insertJposEntriesInDoc( newDoc, entries );

        insertDateSavedComment( newDoc );

        OutputFormat outFormat = new OutputFormat( "xml", "UTF-8", true );

        outFormat.setStandalone( false );
        outFormat.setIndenting( true );
        outFormat.setIndent( 4 );
        outFormat.setPreserveSpace( true );
        outFormat.setLineWidth( 0 );

        insertDTDInfo( newDoc, outFormat );

        PrintWriter outWriter = null;
		try
		{
			outWriter = new PrintWriter
			(new BufferedWriter(new OutputStreamWriter(os, "UTF-8")));
		}
		catch (UnsupportedEncodingException ex)
		{ tracerPrintln("Error making PrintWriter.  Error message = " + ex.getMessage()); }

		if( outWriter != null )
		{
			XMLSerializer xmlSerializer = new XMLSerializer( outWriter, outFormat );
			xmlSerializer.serialize( newDoc );
		}
    }

    /**
     * Appends the <creation> element to the document
     * @param doc the XML Document object
     * @param jposEntryElement the <JposEntryElement> XML Element object
     * @param jposEntry the JposEntry object
     * @since 1.2 (NY 2K meeting)
     */
    protected void appendCreationElement( Document doc, Element jposEntryElement,
                                          JposEntry jposEntry )
    {
        jposEntryElement.appendChild( doc.createTextNode( "    " + "    " ) );

        Element creationElement = doc.createElement( "creation" );

        Attr factoryClassAttr = doc.createAttribute( "factoryClass" );
        Attr serviceClassAttr = doc.createAttribute( "serviceClass" );

        factoryClassAttr.setValue( (String)jposEntry.getPropertyValue( "serviceInstanceFactoryClass" ) );
        serviceClassAttr.setValue( (String)jposEntry.getPropertyValue( "serviceClass" ) );

        creationElement.setAttributeNode( factoryClassAttr );
        creationElement.setAttributeNode( serviceClassAttr );

        jposEntryElement.appendChild( creationElement );
        jposEntryElement.appendChild( doc.createTextNode( "\n" ) );
    }

    /**
     * Appends the <vendor> element to the document
     * @param doc the XML Document object
     * @param jposEntryElement the <JposEntryElement> XML Element object
     * @param jposEntry the JposEntry object
     * @since 1.2 (NY 2K meeting)
     */
    protected void appendVendorElement( Document doc, Element jposEntryElement,
                                        JposEntry jposEntry )
    {
        jposEntryElement.appendChild( doc.createTextNode( "    " + "    " ) );

        Element vendorElement = doc.createElement( "vendor" );

        Attr nameAttr = doc.createAttribute( "name" );
        Attr urlAttr = doc.createAttribute( "url" );

        nameAttr.setValue( (String)jposEntry.getPropertyValue( "vendorName" ) );
        urlAttr.setValue( (String)jposEntry.getPropertyValue( "vendorURL" ) );

        vendorElement.setAttributeNode( nameAttr );
        vendorElement.setAttributeNode( urlAttr );

        jposEntryElement.appendChild( vendorElement );
        jposEntryElement.appendChild( doc.createTextNode( "\n" ) );
    }

    /**
     * Appends the <jpos> element to the document
     * @param doc the XML Document object
     * @param jposEntryElement the <JposEntryElement> XML Element object
     * @param jposEntry the JposEntry object
     * @since 1.2 (NY 2K meeting)
     */
    protected void appendJposElement( Document doc, Element jposEntryElement,
                                      JposEntry jposEntry )
    {
        jposEntryElement.appendChild( doc.createTextNode( "    " + "    " ) );

        Element jposElement = doc.createElement( "jpos" );

        Attr versionAttr = doc.createAttribute( "version" );
        Attr categoryAttr = doc.createAttribute( "category" );

        versionAttr.setValue( (String)jposEntry.getPropertyValue( "jposVersion" ) );
        categoryAttr.setValue( (String)jposEntry.getPropertyValue( "deviceCategory" ) );

        jposElement.setAttributeNode( versionAttr );
        jposElement.setAttributeNode( categoryAttr );

        jposEntryElement.appendChild( jposElement );
        jposEntryElement.appendChild( doc.createTextNode( "\n" ) );
    }

    /**
     * Appends the <product> element to the document
     * @param doc the XML Document object
     * @param jposEntryElement the <JposEntryElement> XML Element object
     * @param jposEntry the JposEntry object
     * @since 1.2 (NY 2K meeting)
     */
    protected void appendProductElement( Document doc, Element jposEntryElement,
                                         JposEntry jposEntry )
    {
        jposEntryElement.appendChild( doc.createTextNode( "    " + "    " ) );

        Element productElement = doc.createElement( "product" );

        Attr nameAttr = doc.createAttribute( "name" );
        Attr descriptionAttr = doc.createAttribute( "description" );
        Attr urlAttr = doc.createAttribute( "url" );

        nameAttr.setValue( (String)jposEntry.getPropertyValue( "productName" ) );
        descriptionAttr.setValue( (String)jposEntry.getPropertyValue( "productDescription" ) );
        urlAttr.setValue( (String)jposEntry.getPropertyValue( "productURL" ) );

        productElement.setAttributeNode( nameAttr );
        productElement.setAttributeNode( descriptionAttr );
        productElement.setAttributeNode( urlAttr );

        jposEntryElement.appendChild( productElement );
        jposEntryElement.appendChild( doc.createTextNode( "\n" ) );
    }

    /**
     * Appends the <prop> element to the document
     * @param doc the XML Document object
     * @param jposEntryElement the <JposEntryElement> XML Element object
     * @param propName the property name
     * @param propValue the property value
     * @since 1.2 (NY 2K meeting)
     */
    protected void appendPropElement( Document doc, Element jposEntryElement,
                                      String propName, Object propValue )
    {
        jposEntryElement.appendChild( doc.createTextNode( "    " + "    " ) );

        Element propElement = doc.createElement( "prop" );

        Attr nameAttr = doc.createAttribute( "name" );
        Attr valueAttr = doc.createAttribute( "value" );
        Attr typeAttr = doc.createAttribute( "type" );

        nameAttr.setValue( propName );
        valueAttr.setValue( propValue.toString() );
        typeAttr.setValue( JposEntryUtility.shortClassName( propValue.getClass() ) );

        propElement.setAttributeNode( nameAttr );
        propElement.setAttributeNode( valueAttr );
        propElement.setAttributeNode( typeAttr );

        jposEntryElement.appendChild( propElement );
        jposEntryElement.appendChild( doc.createTextNode( "\n" ) );
    }

    /**
     * Appends non-required properties name and value
     * @param doc the XML Document object
     * @param jposEntryElement the <JposEntryElement> XML Element object
     * @param jposEntry the JposEntry object
     * @since 1.2 (NY 2K meeting)
     */
    protected void appendPropElements( Document doc, Element jposEntryElement,
                                      JposEntry jposEntry )
    {
        jposEntryElement.appendChild( doc.createTextNode( "\n" + "    " + "    " ) );
        jposEntryElement.appendChild( doc.createComment( "Other non JavaPOS required property (mostly vendor properties and bus specific properties i.e. RS232 )" ) );
        jposEntryElement.appendChild( doc.createTextNode( "\n" ) );

        Enumeration props = jposEntry.getPropertyNames();

        while( props.hasMoreElements() )
        {
            String propName = (String)props.nextElement();

            if( !JposEntryUtility.isRequiredPropName( propName ) )
                appendPropElement( doc, jposEntryElement, propName, jposEntry.getPropertyValue( propName ) );
        }
    }

    /**
     * Insert the <JposEntryElement> in the XML document object
     * @param doc the XML Document object
     * @param jposEntryElement the <JposEntryElement> XML Element object
     * @param jposEntry the JposEntry object
     * @since 1.2 (NY 2K meeting)
     */
    protected void insertJposEntryInDoc( Document doc, Element jposEntryElement,
                                         JposEntry jposEntry )
    {
        appendCreationElement( doc, jposEntryElement, jposEntry );
        appendVendorElement( doc, jposEntryElement, jposEntry );
        appendJposElement( doc, jposEntryElement, jposEntry );
        appendProductElement( doc, jposEntryElement, jposEntry );
        appendPropElements( doc, jposEntryElement, jposEntry );

        doc.getDocumentElement().appendChild( doc.createTextNode( "\n" + "    " ) );
        doc.getDocumentElement().appendChild( jposEntryElement );
        doc.getDocumentElement().appendChild( doc.createTextNode( "\n" + "    " ) );
    }

    /**
     * Insert an Enumeration of <JposEntryElement> objects in the XML document
     * @param doc the XML Document object
     * @param entries an Enumeration of JposEntry objects
     * @since 1.2 (NY 2K meeting)
     */
    protected void insertJposEntriesInDoc( Document doc, Enumeration entries )
    {
        while( entries.hasMoreElements() )
        {
            JposEntry jposEntry = (JposEntry)entries.nextElement();

            if( JposEntryUtility.isValidJposEntry( jposEntry ) )
            {
                doc.getDocumentElement().appendChild( doc.createTextNode( "\n" + "    " ) );

                Element jposEntryElement = doc.createElement( "JposEntry" );

                Attr logicalNameAttr = doc.createAttribute( "logicalName" );
                logicalNameAttr.setValue( (String)jposEntry.getPropertyValue( "logicalName" ) );

                jposEntryElement.setAttributeNode( logicalNameAttr ); 

                jposEntryElement.appendChild( doc.createTextNode( "\n" ) );

                insertJposEntryInDoc( doc, jposEntryElement, jposEntry );
            }
        }
    }

    /**
     * @return an empty XML Document object
     * @since 1.2 (NY 2K meeting)
     */
    protected Document createEmptyDocument()
    {
        DOMImplementationImpl domImpImpl = (DOMImplementationImpl)DOMImplementationImpl.getDOMImplementation();
        DocumentType docType = domImpImpl.createDocumentType( "JposEntries", "-//JavaPOS//DTD//EN", "jpos/res/jcl.dtd" );
        Document doc = domImpImpl.createDocument( null, "JposEntries", docType );

        return doc;
    }

    /**
     * Inset DTD information in the XML Document object
     * @param doc the XML Document object
     * @param outFormat the OuputFormat object
     * @exception java.lang.Exception in case something goes wrong
     * @since 1.2 (NY 2K meeting)
     */
    protected void insertDTDInfo( Document doc, OutputFormat outFormat ) throws Exception
    {
        String publicId = outFormat.whichDoctypePublic( doc );
        String systemId = outFormat.whichDoctypeSystem( doc );

        outFormat.setDoctype( "JposEntries", "jpos/res/jcl.dtd" );
    }

    /**
     * Inserts date and info saved in the XML Document object
     * @param document the XML Document object
     * @exception java.lang.Exception in case something goes wrong
     * @since 1.2 (NY 2K meeting)
     */
    protected void insertDateSavedComment( Document document ) throws Exception
    {
        String dateString = DateFormat.getInstance().format( new Date( System.currentTimeMillis() ) );

        String commentString = "Saved by JavaPOS jpos.config/loader (JCL) version " + 
                               Version.getVersionString() + " on " + dateString;

        Comment comment = document.createComment( commentString );

        document.getDocumentElement().insertBefore( comment, document.getDocumentElement().getFirstChild() );

        document.getDocumentElement().insertBefore(  document.createTextNode( "\n" ), comment );
        document.getDocumentElement().appendChild( document.createTextNode( "\n" ) );
    }

    //--------------------------------------------------------------------------
    // Instance variables
    //

    private String xmlFileName = "";

    private DOMParser domParser = new DOMParser();

    private boolean createdTempDTD = false;
    private boolean createdTempDir = false;

    //--------------------------------------------------------------------------
    // Public constants
    //

	public static final String DTD_JPOS_FILE_PATH = "jpos";
	public static final String DTD_JPOS_RES_FILE_PATH = "jpos" + File.separator + "res";

    public static final String DTD_JAR_FILE_NAME = "jpos/res/jcl.dtd";
    public static final String DTD_FILE_PATH = "jpos" + File.separator + "res";
    public static final String DTD_FILE_NAME = DTD_FILE_PATH + File.separator + "jcl.dtd";

	/**
	 * The default name for the XercesRegPopulator
	 * @since 1.3 (Washington DC 2001 meeting)
	 */
	public static final String XERCES_REG_POPULATOR_NAME_STRING = "JCL XML Entries Populator";
}
